﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  [Item(Name = "MeanProd", Description = "Product of mean functions for Gaussian processes.")]
  public class MeanProd : Item, IMeanFunction {
    [Storable]
    private ItemList<IMeanFunction> factors;

    [Storable]
    private int numberOfVariables;

    public ItemList<IMeanFunction> Factors {
      get { return factors; }
    }

    public int GetNumberOfParameters(int numberOfVariables) {
      this.numberOfVariables = numberOfVariables;
      return factors.Select(t => t.GetNumberOfParameters(numberOfVariables)).Sum();
    }

    [StorableConstructor]
    protected MeanProd(bool deserializing)
      : base(deserializing) {
    }

    protected MeanProd(MeanProd original, Cloner cloner)
      : base(original, cloner) {
      this.factors = cloner.Clone(original.factors);
      this.numberOfVariables = original.numberOfVariables;
    }

    public MeanProd() {
      this.factors = new ItemList<IMeanFunction>();
    }

    public void SetParameter(double[] hyp) {
      int offset = 0;
      foreach (var t in factors) {
        var numberOfParameters = t.GetNumberOfParameters(numberOfVariables);
        t.SetParameter(hyp.Skip(offset).Take(numberOfParameters).ToArray());
        offset += numberOfParameters;
      }
    }

    public void SetData(double[,] x) {
      foreach (var t in factors) t.SetData(x);
    }

    public double[] GetMean(double[,] x) {
      var res = factors.First().GetMean(x);
      foreach (var t in factors.Skip(1)) {
        var a = t.GetMean(x);
        for (int i = 0; i < res.Length; i++) res[i] *= a[i];
      }
      return res;
    }

    public double[] GetGradients(int k, double[,] x) {
      double[] res = Enumerable.Repeat(1.0, x.GetLength(0)).ToArray();
      foreach (var f in factors) {
        var numParam = f.GetNumberOfParameters(numberOfVariables);
        if (k >= 0 && k < numParam) {
          // multiply gradient
          var g = f.GetGradients(k, x);
          for (int i = 0; i < res.Length; i++) res[i] *= g[i];
          k -= numParam;
        } else {
          // multiply mean
          var m = f.GetMean(x);
          for (int i = 0; i < res.Length; i++) res[i] *= m[i];
          k -= numParam;
        }
      }
      return res;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new MeanProd(this, cloner);
    }
  }
}
